# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
import urllib2
import datetime
import subprocess
from os.path import join, exists
from gettext import gettext as _

import dbus
import dbus.service
import gobject

from sugar_client import util, progress, env
from sugar_client.util import enforce


INTERFACE = 'org.sugarlabs.client.Backup'
NAME = 'org.sugarlabs.client.Backup'
PATH = '/org/sugarlabs/client/Backup'

_RSYNC = '/usr/bin/rsync'
# Exponential timeout between checks for backup availability
_BACKUP_SERVER_TIMEOUT = lambda a_try: 60 * 2 ** (a_try - 1)


class Service(env.SessionService):

    def __init__(self):
        env.SessionService.__init__(self, INTERFACE, NAME, PATH)
        self._process = None
        self._locked = False

    @dbus.service.method(INTERFACE,
            in_signature='si', out_signature='b')
    def Backup(self, url, tries):
        """Backup datastore to the server.

        The function will ask backup server if it is possible to start backup
        process. If server will reply with reject, function will try `tries`
        times with exponentially growth timeout and will raise `RuntimeError`
        eventually.

        :param url:
            rsync url to backup data to
        :param tries:
            number of tries before raising a `RuntimeError`,
            makes sense only for non-local-fs `url`s

        """
        if self._locked:
            logging.debug('Another action is in progress')
            return False

        self._locked = True
        progress.info(_('Start Journal backup'))

        def backup(url):
            progress.info(_('Backup Journal to %s'), url)
            self._rsync(env.profile_path('datastore'), url, None)

        if not url:
            enforce(env.backup_url.value,
                    _('Registeration was not processed, no backup url'))
            if ':' in env.backup_url.value:
                # Use url that was gotten after registeration
                enforce(env.backup_availability_url.value,
                        _('Registeration was not processed, ' \
                                'no backup availability url'))
                gobject.idle_add(self._check_backup_server,
                        env.backup_availability_url.value, 1, tries or 7,
                        lambda: backup(env.backup_url.value))
            else:
                backup(env.backup_url.value)
        else:
            backup(url)

        return True

    @dbus.service.method(INTERFACE,
            in_signature='si', out_signature='b')
    def Restore(self, url, date):
        if self._locked:
            logging.debug('Another action is in progress')
            return False

        self._locked = True
        progress.info(_('Start Journal restore'))

        if not url:
            url = env.backup_url.value
            if date:
                url += '/' + datetime.date.fromtimestamp.strftime('%Y-%m-%d')
        else:
            enforce(date == 0,
                    _('Backup date might be used only for remote restore'))

        progress.info(_('Restore Journal from %s'), url)
        dst_path = env.profile_path('datastore')

        def finish_cb():
            if exists(join(dst_path, 'index_updated')):
                os.unlink(join(dst_path, 'index_updated'))
            util.rmtree(join(dst_path, 'index'))
            self._finished(0)

        self._rsync(url, dst_path, finish_cb)

        return True

    @dbus.service.method(INTERFACE,
            in_signature='', out_signature='')
    def Cancel(self):
        if not self._locked:
            return
        logging.info(_('Cancel current action'))
        if self._process is None:
            # The process is in timeout, just reset _locked flag
            # to discard it on awake
            self._finished(1)
        else:
            self._process.kill()

    @dbus.service.signal(INTERFACE,
            signature='u')
    def Finished(self, exit_status):
        pass

    @property
    def Locked(self):
        return self._locked

    def _rsync(self, src, dst, finish_cb):
        # Add a trailing slash to ensure that we don't generate dst subdir
        src = src.rstrip(os.sep) + os.sep

        ssh_cmd = '/usr/bin/ssh -F /dev/null -o "PasswordAuthentication no" ' \
                '-o "StrictHostKeyChecking no" -i "%s"' % \
                env.profile_path('owner.key')
        rsync_cmd = [_RSYNC, '-z', '-rlt', '--delete', '--timeout=160',
                '--exclude=/index', '--exclude=/index_updated',
                '--partial-dir=.rsync-partial', '--out-format=%i',
                '-e', ssh_cmd, src, dst]
        logging.debug('Backup rsync command: %r', rsync_cmd)

        progress.reset()
        # At first, calculate number of transferes to have total progress
        self._popen(rsync_cmd + ['--dry-run'], progress.increment_total,
                # Then, do a real transfer
                lambda: self._popen(rsync_cmd, progress.increment, finish_cb))

    def _popen(self, cmd, progress_cb, finish_cb):
        self._process = subprocess.Popen(cmd,
                stderr=subprocess.PIPE, stdout=subprocess.PIPE)

        def io_cb(fd, condition):
            if condition & gobject.IO_IN:
                fd.readline().rstrip()
                progress_cb()
            if condition & gobject.IO_HUP:
                self._process.wait()
                returncode = self._process.returncode
                stderr = self._process.stderr.read().strip()
                self._process = None
                if returncode == 0:
                    if finish_cb is None:
                        self._finished(0)
                    else:
                        finish_cb()
                else:
                    progress.info(_('Rsync failed with code %s: %s'),
                            returncode, stderr)
                    self._finished(returncode)
                return False
            else:
                return True

        gobject.io_add_watch(self._process.stdout,
                gobject.IO_IN | gobject.IO_HUP, io_cb)

    def _check_backup_server(self, url, a_try, tries, finish_cb):
        if not self._locked:
            # Canceled
            return False

        try:
            message = _('Check for backup availability using %s') % url
            if a_try > 1:
                message += ' (%d)' % a_try
            progress.info(message)

            logging.debug('Ping backup server on %s %r %r', url, a_try, tries)
            urllib2.urlopen(url).read()
            logging.debug('Got accept from backup server on %s', url)
            finish_cb()

        except urllib2.HTTPError, error:
            if error.code == 503:
                if a_try < tries:
                    timeout = _BACKUP_SERVER_TIMEOUT(a_try)
                    logging.debug('Backoff timeout for %d seconds before ' \
                            'trying backup once more', timeout)
                    gobject.timeout_add_seconds(timeout,
                            self._check_backup_server, url, a_try + 1, tries,
                            finish_cb)
                else:
                    progress.info(_('Backup server is too busy'))
                    self._finished(1)
            else:
                progress.info(_('Backup server failed: %s'), error)
                self._finished(1)
        except urllib2.URLError, error:
            progress.info(_('Cannot connect to backup server: %s') % error)
            self._finished(1)
        except Exception:
            progress.info(_('Internal fail while connecting to backup server'))
            self._finished(1)
            raise

        return False

    def _finished(self, exit_status):
        # Need _finished() to set _locked before calling Finished()
        self._locked = False
        self.Finished(exit_status)
