# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pwd
import fcntl
import errno
import logging
from os.path import join, exists, dirname
from gettext import gettext as _

import dbus.service

from sugar_client import util, progress
from sugar_client.util import enforce


VERSION = '1.0'

_XO_SERIAL_PATH = '/ofw/mfg-data/SN'
_XO_UUID_PATH = '/ofw/mfg-data/U#'
_DBUS_UUID_PATH = '/var/lib/dbus/machine-id'

debug = util.Option(
        _('debug logging level; multiple argument'),
        default=0, type_cast=int, short_option='-D', action='count')

user = util.Option(
        _('system user of sugar client, needs to be set if program is run ' \
                'by root user'),
        default='olpc')

profile = util.Option(
        _('sugar profile name, needs to be set if program is run ' \
                'by root user'),
        default='default')

jabber = util.Option(
        _('jabber server for Sugar collaboraiton'))

backup_url = util.Option(
        _('rsync url to backup/restore Journal'))

backup_availability_url = util.Option(
        _('url to ask server before making backup'))

pms_config = util.Option(
        _('path to the Package Management System configuration file that ' \
                'will be used for upgrade command'))


def init_logger(**kwargs):
    if not debug.value:
        level = logging.WARNING
    elif debug.value == 1:
        level = logging.INFO
    else:
        level = logging.DEBUG
    logging.basicConfig(level=level, **kwargs)


def var_path(filename):
    """Full path to the file in an analog of /var directory.

    :param filename:
        filename without dirs to avoid permission clashing between
        several users (root and sugar one)
    :returns:
        full path with directory part existed

    """
    enforce(os.sep not in filename)

    # The reason to use /tmp is that OLPC OS keeps /var on tmpfs and something
    # like /var/run/sugar-client needs to be recreated by root (to make it
    # writtable for non-root user) on every boot.
    result = join(os.environ.get('TMPDIR', '/tmp'), filename)

    if os.geteuid() == 0 and not exists(result):
        pw_user = pwd.getpwnam(user.value)
        # The file itself should exist to be owned by `user`
        file(result, 'w').close()
        os.chown(result, pw_user.pw_uid, pw_user.pw_gid)

    return result


def profile_path(*args):
    """Path to the sugar profile directory.

    Even for root user this function will return valid path within `user`'s
    home. In that case, the full path will exist and created by `user` user.

    :param args:
        path parts that will be added to the resulting path
    :returns:
        full path with directory part existed

    """
    pw_user = pwd.getpwnam(user.value)
    result = join(pw_user.pw_dir, '.sugar', profile.value, *args)

    if not exists(dirname(result)):
        if os.geteuid():
            enforce(os.geteuid() == pw_user.pw_uid,
                    _('Current system user is not the same as --user'))
            if not exists(dirname(result)):
                os.makedirs(dirname(result))
        else:
            # Create directories and make it owned by `user`
            path = os.sep
            for i in dirname(result).split(os.sep)[1:]:
                path = join(path, i)
                if not exists(path):
                    os.makedirs(path)
                    os.chown(path, pw_user.pw_uid, pw_user.pw_gid)

    if os.geteuid() == 0 and not exists(result):
        # The file itself should exist as well to be owned by `user`
        file(result, 'w').close()
        os.chown(result, pw_user.pw_uid, pw_user.pw_gid)

    return result


def log_path():
    return profile_path('logs', 'sugar-client.log')


def serial():
    """Machine Serial Number value.

    This function makes sense only for XO laptops. In non-XO environments,
    it returns constant, `SHF00000000`, value.

    :returns:
        machine serial number

    """
    if exists(_XO_SERIAL_PATH):
        return _read_XO_value(_XO_SERIAL_PATH)
    else:
        return 'SHF00000000'


def uuid():
    """Machine UUID value.

    On a XO, function will return value flashed to XO. Otherwise, it will be
    a machine ID generated by DBus.

    :returns:
        string with UUID value

    """
    if exists(_XO_UUID_PATH):
        return _read_XO_value(_XO_UUID_PATH)
    else:
        return file(_DBUS_UUID_PATH).read().strip()


def nick():
    return gconf_get('/desktop/sugar/user/nick')


def flock(name):
    """Make a file lock.

    :param name:
        a string that will be used in filename
    :returns:
        file object that was locked; otherwise, RuntimeError will be raisen

    """
    lock_path = util.TempFilePath(var_path('sugar-client-%s.lock' % name))
    lock = file(lock_path, 'a+')

    try:
        fcntl.flock(lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
        return lock
    except IOError, error:
        lock.close()
        if error.errno in [errno.EACCES, errno.EAGAIN]:
            raise RuntimeError(_('Another %s is in progress') % name)
        raise


def http_proxy():
    host = gconf_get('/system/http_proxy/host')
    port = gconf_get('/system/http_proxy/port')
    if gconf_get('/system/http_proxy/use_http_proxy') != 'true':
        return None
    elif gconf_get('/system/http_proxy/use_authentication') == 'true':
        auth_user = gconf_get('/system/http_proxy/authentication_user')
        auth_pswd = gconf_get('/system/http_proxy/authentication_password')
        return 'http://%s:%s@%s:%s' % (auth_user, auth_pswd, host, port)
    else:
        return 'http://%s:%s' % (host, port)


def gconf_get(key):
    """Read GConf value.

    The purpose to use subprocesse is that gconf python module fails being
    started by su'ed system user. This function might be called in python
    session started by root system user, in that case it will su to --user
    before launching gconftool-2.

    :param key:
        GConf key to read
    :returns:
        GConf key value in string notation, i.e., how gconftool-2 returned

    """
    cmd = '/usr/bin/gconftool-2 -g %s' % key
    if os.geteuid() == 0:
        cmd = 'su %s -c "%s"' % (user.value, cmd)
    output = util.assert_call(cmd, shell=True, raw=True)
    # Skip the last '\n'
    return output[:-1]


def gconf_set(key, value):
    """Set GConf string value.

    The purpose to use subprocesse is that gconf python module fails being
    started by su'ed system user. This function might be called in python
    session started by root system user, in that case it will su to --user
    before launching gconftool-2.

    :param key:
        GConf key, of string tupe, to set
    :param value:
        GConf value to set

    """
    cmd = "/usr/bin/gconftool-2 -s %s --type string '%s'" % (key, value)
    if os.geteuid() == 0:
        cmd = 'su %s -c "%s"' % (user.value, cmd)
    util.assert_call(cmd, shell=True, raw=True)


class Service(dbus.service.Object):

    def __init__(self, bus, interface, name, path):
        bus_name = dbus.service.BusName(name, bus=bus)
        dbus.service.Object.__init__(self, bus_name, path)

        self.interface = interface
        self.name = name
        self.path = path

        progress.status_changed_cb = \
                lambda: self.PropertiesChanged(
                        self.interface, {'Status': progress.status}, [])
        progress.percent_changed_cb = \
                lambda: self.PropertiesChanged(
                        self.interface, {'Progress': progress.percent}, [])

    @property
    def Progress(self):
        return progress.percent or 0

    @property
    def Status(self):
        return progress.status or ''

    @dbus.service.method(dbus.PROPERTIES_IFACE,
            in_signature='ss', out_signature='v')
    def Get(self, interface, prop):
        enforce(interface == self.interface,
                _('Interface %s is not supported'), interface)
        enforce(hasattr(self, prop) and \
                isinstance(getattr(self.__class__, prop), property),
                _('Unsupported %s property'), prop)
        return getattr(self, prop)

    @dbus.service.method(dbus.PROPERTIES_IFACE,
            in_signature='s', out_signature='a{sv}')
    def GetAll(self, interface):
        enforce(interface == self.interface,
                _('Interface %s is not supported'), interface)
        result = {}
        for i in dir(self.__class__):
            if isinstance(getattr(self.__class__, i), property):
                result[i] = getattr(self, i)
        return result

    @dbus.service.method(dbus.PROPERTIES_IFACE,
            in_signature='ssv')
    def Set(self, interface, prop, value):
        enforce(interface == self.interface,
                _('Interface %s is not supported'), interface)
        enforce(hasattr(self, prop) and \
                isinstance(getattr(self.__class__, prop), property),
                _('Unsupported %s property'), prop)
        setattr(self, prop, value)
        self.PropertiesChanged(interface, {prop: value}, [])

    @dbus.service.signal(dbus.PROPERTIES_IFACE,
            signature='sa{sv}as')
    def PropertiesChanged(self, interface, changed_props, invalidated_props):
        pass


class SessionService(Service):

    def __init__(self, interface, name, path):
        Service.__init__(self, dbus.SessionBus(), interface, name, path)


class SystemService(Service):

    def __init__(self, interface, name, path):
        Service.__init__(self, dbus.SystemBus(), interface, name, path)


def _read_XO_value(path):
    return file(path).read().rstrip('\0\n')


util.Option.seek('main')
util.Command.seek('main')
