# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import socket
import logging
import xmlrpclib
from ConfigParser import ConfigParser
from os.path import exists
from gettext import gettext as _

import dbus
import dbus.service

from sugar_client import util, progress, env
from sugar_client.util import enforce


INTERFACE = 'org.sugarlabs.client.Server'
NAME = 'org.sugarlabs.client.Server'
PATH = '/org/sugarlabs/client/Server'

_SOCKET_TIMEOUT = 8
_SUGAR_PUBKEY_PREFIX = 'ssh-dss '
_DEFAULT_SCHOOLSERVER_URL = 'http://schoolserver:8080/'


class Service(env.SessionService):

    def __init__(self):
        env.SessionService.__init__(self, INTERFACE, NAME, PATH)

    @dbus.service.method(INTERFACE,
            in_signature='s', out_signature='')
    def Register(self, url):
        if not url:
            url = _DEFAULT_SCHOOLSERVER_URL

        progress.info(_('Registring on %s'), url)

        pubkey_path = env.profile_path('owner.key.pub')
        enforce(exists(pubkey_path),
                _('Sugar session was never started, no pubkey'))
        pubkey = _read_pubkey(pubkey_path)

        try:
            socket.setdefaulttimeout(_SOCKET_TIMEOUT)
            server = xmlrpclib.ServerProxy(url)
            reply = server.register(env.serial(), env.nick(), env.uuid(),
                    pubkey)
        finally:
            socket.setdefaulttimeout(None)

        logging.debug('Got registration reply: %r', reply)

        enforce(reply.get('success') == 'OK',
                _('Registration on a server using "%s" url was failed: %s'),
                url, reply.get('error', _('Unknown error')))
        reply = reply.get('config', {})

        config = ConfigParser()
        config.add_section('main')
        for key, value in reply.items():
            config.set('main', key, value)
            if key in util.Option.items:
                util.Option.items[key].value = value
        config_file = file(env.profile_path('sugar-client.conf'), 'w')
        config.write(config_file)
        config_file.close()

        env.gconf_set('/desktop/sugar/collaboration/jabber_server',
                env.jabber.value or reply.get('jabber'))

    @property
    def Registered(self):
        return exists(env.profile_path('sugar-client.conf'))


def _read_pubkey(path):
    for line in file(path).readlines():
        line = line.strip()
        if line.startswith(_SUGAR_PUBKEY_PREFIX):
            return line[len(_SUGAR_PUBKEY_PREFIX):]
    raise RuntimeError(_('Valid SSH public key was not found in %s') % path)
