# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import datetime
from os.path import exists
from gettext import gettext as _

import dbus
import dbus.service
import gobject

from sugar_client import progress, env, pms, util
from sugar_client.util import enforce


INTERFACE = 'org.sugarlabs.client.System'
NAME = 'org.sugarlabs.client.System'
PATH = '/org/sugarlabs/client/System'

# Backwards compatibility with dextrose-updater
_INTERFACE_COMPAT = 'org.sugarlabs.system'
_PATH_COMPAT = '/org/sugarlabs/system'

_STAMP_PATH = '/tmp/sugar-client-upgrade'


class Service(env.SystemService):

    def __init__(self):
        env.SystemService.__init__(self, INTERFACE, NAME, PATH)
        self._compat = _ServiceCompat()

        environ = os.environ.copy()
        environ['http_proxy'] = env.http_proxy() or ''

        self._pms_config = None
        if env.pms_config.value:
            self._pms_config = util.TempFilePath(text=env.pms_config.value)
        self._pms = pms.get(self._pms_config, environ)

    def remove_from_connection(self, connection=None, path=None):
        self._compat.remove_from_connection(connection, path)
        env.SystemService.remove_from_connection(self, connection, path)

    @dbus.service.method(INTERFACE,
            in_signature='b', out_signature='')
    def Update(self, force):
        enforce(env.pms_config.value, _('The option pms-config is not set'))

        if not force and _get_stamp() == datetime.date.today():
            progress.info(_('Daily check for upgrade already happened, ' \
                    'pass --force to ignore it'))
            gobject.idle_add(lambda: self.Finished(0))
            return

        def update():
            try:
                self._update()
            except Exception, error:
                progress.info(_('Error while procesing system update: %s'),
                        error)
                self.Finished(1)
            else:
                self.Finished(0)

        gobject.idle_add(update)

    @dbus.service.signal(INTERFACE,
            signature='u')
    def Finished(self, exit_status):
        pass

    @dbus.service.signal(INTERFACE, signature='')
    def Reboot(self):
        self._compat.Reboot()

    @dbus.service.signal(INTERFACE, signature='')
    def Relogin(self):
        self._compat.Relogin()

    @property
    def LastUpdate(self):
        return self._pms.last_updated()

    def _update(self):
        # Let self._pms make sure that environment is ready to start updating
        self._pms.cleanup()

        to_update = set(self._pms.to_update())
        if not to_update:
            _set_stamp()
            progress.info(_('No packages to update'))
            return

        progress.info(_('Update %s package(s)'), ', '.join(to_update))
        self._pms.update()
        _set_stamp()

        after_update = set(self._pms.to_update())
        if after_update:
            progress.info(_('Not updated packages: %s'), ', '.join(to_update))

        if (to_update & set(self._pms.reboot_packages)) != \
                (after_update & set(self._pms.reboot_packages)):
            progress.info(_('Need to reboot'))
            self.Reboot()

        if (to_update & set(self._pms.relogin_packages)) != \
                (after_update & set(self._pms.relogin_packages)):
            progress.info(_('Need to relogin'))
            self.Relogin()


class _ServiceCompat(dbus.service.Object):

    def __init__(self):
        dbus.service.Object.__init__(self, dbus.SystemBus(), _PATH_COMPAT)

    @dbus.service.signal(_INTERFACE_COMPAT, signature='')
    def Reboot(self):
        pass

    @dbus.service.signal(_INTERFACE_COMPAT, signature='')
    def Relogin(self):
        pass


def _get_stamp():
    if not exists(_STAMP_PATH):
        return None
    return datetime.date.fromtimestamp(os.stat(_STAMP_PATH).st_mtime)


def _set_stamp():
    file(_STAMP_PATH, 'w').close()
